/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.generator.common.utils.SourceModelElementFactory.createStaticContentSourceUnit;
import static org.fortiss.af3.platform.raspberry.utils.RaspberryUtils.getEcuForComponent;
import static org.fortiss.af3.platform.raspberry.utils.RaspberryUtils.getTargetComponent;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.findElementById;

import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.PortSpecification;
import org.fortiss.af3.deployment.model.ComponentAllocation;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.generator.common.model.source.SourcePackage;
import org.fortiss.af3.platform.language.executable.TransmissionUnitExecutableBase;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.TransmissionUnit;
import org.fortiss.af3.platform.model.impl.PlatformConnectorUnitImpl;
import org.fortiss.af3.platform.raspberry.generated.files.RaspberryFileGeneration;
import org.fortiss.af3.platform.raspberry.model.annotation.PeripheralAddress;
import org.fortiss.af3.platform.raspberry.model.annotation.PinNumber;
import org.fortiss.af3.platform.raspberry.model.impl.ActuatorOutImpl;
import org.fortiss.af3.platform.raspberry.model.impl.SensorInImpl;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;

/**
 * 
 * @author chaudhary
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class BusExecutable extends TransmissionUnitExecutableBase<TransmissionUnit> {

	/**
	 * @param modelElement
	 */
	public BusExecutable(TransmissionUnit modelElement) {
		super(modelElement);
		// TODO Auto-generated constructor stub
	}

	/** {@inheritDoc} */
	@Override
	public void createAllFiles(SourcePackage pack, Map<ExecutionUnit, List<Port>> euPortsPair,
			Map<ExecutionUnit, List<Component>> euComponentList, Component topComponent,
			int deploymentID) {
		findBusSignals(deploymentID, topComponent, pack);
	}

	/**
	 * Finds Information about Signals that are transmitted over the Bus and form fields of the
	 * message.
	 */
	public void findBusSignals(int deploymentID, Component topComponent, SourcePackage pack) {
		createProtoFile(deploymentID, topComponent, pack);

		createHeaderFile(pack, "sensors");
		createHeaderFile(pack, "actuators");

		createFilesPeripheralDetails(pack);
	}

	/**
	 * Finds Information about Signals that are transmitted over the Bus and form fields of the
	 * message.
	 */
	private void createProtoFile(int deploymentID, Component topComponent, SourcePackage pack) {
		String messageString = "message msg { \n";
		int counter = 0;
		Deployment currentDeployment =
				(Deployment)findElementById(deploymentID, topComponent.eContainer().eContainer());
		for(ComponentAllocation allocation : currentDeployment.getComponentAllocations()) {
			Component c1 = allocation.getComponent();
			Object lastPort = null;
			for(Object port : c1.getConnectors()) {
				if(port instanceof OutputPort) {
					for(Component c2 : getTargetComponent((OutputPort)port)) {
						if(getEcuForComponent(c1, currentDeployment) != getEcuForComponent(c2,
								currentDeployment) &&
								getEcuForComponent(c2, currentDeployment) != null) {
							for(Object specs : ((OutputPort)port).getSpecifications()) {
								if(specs instanceof PortSpecification && port != lastPort) {
									lastPort = port;
									counter = counter + 1;
									PortSpecification portSpec = (PortSpecification)specs;
									String prtType;
									switch(portSpec.getType().toString()) {
										case "int":
											prtType = "int32";
											break;
										case "boolean":
											prtType = "bool";
											break;
										default:
											prtType = portSpec.getType().toString();
											break;
									}
									messageString +=
											"\t optional " + prtType + " " +
													((OutputPort)port).getName().toLowerCase() +
													"= " + counter + ";\n";
								}
							}
						}
					}
				}
			}
		}
		messageString += "}";
		createFile(pack, messageString, "message.proto");
	}

	/** Determine Sensors, Actuators their IDs and PinNumbers. */
	private void createFilesPeripheralDetails(SourcePackage pack) {
		String sensorList = "#include \"sensors.h\" \n\n";
		sensorList += "sensors sen[] = \n{ \n";
		String actuatorList = "#include \"actuators.h\" \n\n";
		actuatorList += "actuators act[] = \n{\n";

		String componentsList = "#ifndef __COMPONENTS_H\n";
		componentsList += "#define __COMPONENTS_H\n\n";

		EList<IHierarchicElement> el = modelElement.getContainer().getContainedElements();
		for(IHierarchicElement ce : el) {
			EList<IConnector> conList = ce.getConnectors();
			for(IConnector conn : conList) {
				if(conn instanceof SensorInImpl) {
					sensorList += addSensorSpecsToList((SensorInImpl)conn);
					componentsList +=
							componentDefine("SensorIn", ((PlatformConnectorUnitImpl)conn).getId());
				}
				if(conn instanceof ActuatorOutImpl) {
					actuatorList += addSensorSpecsToList((ActuatorOutImpl)conn);
					componentsList +=
							componentDefine("ActuatorOut",
									((PlatformConnectorUnitImpl)conn).getId());
				}
			}
		}

		sensorList = fixEnding(sensorList);
		actuatorList = fixEnding(actuatorList);

		componentsList += "\n#endif";

		createFile(pack, sensorList, "sensors.c");
		createFile(pack, actuatorList, "actuators.c");
		createFile(pack, componentsList, "components.h");
	}

	private String fixEnding(String s) {
		return s.substring(0, s.length() - ",\n".length()) + "\n};";

	}

	private String componentDefine(String s, int value) {
		return "#define " + s + "_ID_" + value + " " + value + "\n";
	}

	/** Add specification of peripheral to the respective list of Sensors. */
	private String addSensorSpecsToList(PlatformConnectorUnitImpl conn) {
		int id = conn.getId();
		int pinNr = 0;
		int address = 0;
		EList<IModelElementSpecification> specs = conn.getSpecifications();
		for(IModelElementSpecification spec : specs) {
			if(spec instanceof PinNumber) {
				pinNr = ((PinNumber)spec).getPinNumber();
			}
			if(spec instanceof PeripheralAddress) {
				address = ((PeripheralAddress)spec).getPeripheralAddress();
			}
		}
		return "\t{" + id + ", " + pinNr + ", " + address + "},\n";
	}

	/** Create Header Files defining structure for actuator. */
	private void createHeaderFile(SourcePackage pack, String type) {
		String content = "#ifndef __" + type.toUpperCase() + "_H \n";
		content += "#define __" + type.toUpperCase() + "_H \n\n";
		content += "struct " + type + "\n{ \n";
		content += "\tint Id;\n\tint pinNr;\n\tint address; \n};\n\n";
		content += "typedef struct " + type + " " + type + ";\n\n";
		content += "extern " + type + " " + type.substring(0, 3) + "[];\n\n";
		content += "#endif \n \n";
		createFile(pack, content, type + ".h");
	}

	/** Create file with a given Name and given text. */
	private void createFile(SourcePackage pack, String text, String name) {
		RaspberryFileGeneration file = new RaspberryFileGeneration(text);
		file.createInfofile();
		pack.addUnit(createStaticContentSourceUnit(name, file.getContent(), false));
	}
}
